import torch
import torch.nn as nn

from .categorical_tensor import CategoricalTensor

class Module(nn.Module):

    def __init__(self):
        super().__init__()
        self.initialized = False
        self.submodule_index = 0
        self.submodules = nn.ModuleList()

    def start_forward(self):
        if self.submodule_index != 0:
            self.initialized = True
        self.submodule_index = 0

    def make(self, cls, *args, **kwargs):
        if not self.initialized:
            self.submodules.append(cls(*args, **kwargs))
        submod = self.submodules[self.submodule_index]
        self.submodule_index += 1
        return submod

class WrapperModule(Module):

    def __init__(self, *args, **kwargs):
        super().__init__()
        self.args = args
        self.kwargs = kwargs

class LazyLinear(WrapperModule):
    """ torch >=1.13 has this, but wanted to try my hand at implimenting it myself
    (and ensuring that older torch versions work) """

    def forward(self, x):
        self.start_forward()
        in_feats = x.shape[-1]
        return self.make(nn.Linear, in_feats, *self.args, **self.kwargs)(x)


class LazyEmbedding(Module):

    def __init__(self, embedding_dims):
        super().__init__()
        self.embedding_dims = embedding_dims

    def forward(self, x: CategoricalTensor):
        self.start_forward()
        embedding_dims = [self.embedding_dims]*len(x.max_values) if isinstance(self.embedding_dims, int) else self.embedding_dims
        ret = []
        for idx in range(x.shape[-1]):
            max_val = x.max_values[idx]
            ret.append(self.make(nn.Embedding, max_val, embedding_dims[idx])(x.tensor[..., idx]))
        return torch.cat(ret, -1)



from typing import Tuple, Union
import functools
import torch

from torch.utils._pytree import tree_map

# https://pytorch.org/docs/stable/notes/extending.html

# https://colab.research.google.com/drive/1MLeSCMrc6a5Yf_6uAh0bH-IPf5c_BTQf#scrollTo=dM0sl2X6epBX


HANDLED_FUNCTIONS = {}
# assume max values apply to the last axis of the tensor
class CategoricalTensor(torch.Tensor):
    @staticmethod
    def __new__(cls, tensor: torch.Tensor, max_values: Tuple[int]):
        return torch.Tensor._make_subclass(cls, tensor.to('meta'))

    def __init__(self, tensor, max_values: Tuple[int]):
        assert tensor.dtype == torch.long
        assert len(max_values) == tensor.size(-1)
        self.tensor = tensor
        self.max_values = max_values

    def __repr__(self):
        return f"{self.__class__.__name__}(tensor={self.tensor}, max_values={self.max_values})"

    # Only difference with above is to add this function.
    @classmethod
    def __torch_function__(cls, func, types, args=(), kwargs=None):
        if kwargs is None:
            kwargs = {}
        if func in HANDLED_FUNCTIONS:
            return HANDLED_FUNCTIONS[func](*args, **kwargs)
        raise NotImplementedError()

def implements(torch_function):
    """Register a torch function override for ScalarTensor"""
    @functools.wraps(torch_function)
    def decorator(func):
        HANDLED_FUNCTIONS[torch_function] = func
        return func
    return decorator

@implements(torch.cat)
def cat(inputs: Tuple[CategoricalTensor, ...], axis: int = -1) -> CategoricalTensor:
    
    assert axis == -1 # for now
    tensor = torch.cat([t.tensor for t in inputs], axis)
    max_values = []
    for t in inputs:
        max_values += t.max_values
    return CategoricalTensor(tensor, max_values)

@implements(torch.stack)
def stack(inputs: Tuple[CategoricalTensor, ...]):
    assert len(inputs) > 0
    max_values = inputs[0].max_values
    for t in inputs:
        assert t.max_values == max_values
    return CategoricalTensor(torch.stack([t.tensor for t in inputs]), max_values)

@implements(torch.Tensor.__getitem__)
def getitem(t: CategoricalTensor, idx: int):
    max_values = t.max_values
    if isinstance(idx, tuple):
        if len(idx) == len(t.shape):
            max_values = t.max_values[idx[-1]]
    else:
        if len(t.shape) == 1:
            max_values = t.max_values[idx]
    if isinstance(max_values, int):
        max_values = [ max_values ]
    return CategoricalTensor(t.tensor[idx], max_values)

@implements(torch.Tensor.shape.__get__)
def get_shape(t: CategoricalTensor):
    return t.tensor.shape

@implements(torch.Tensor.device.__get__)
def get_device(t: CategoricalTensor):
    return t.tensor.device

@implements(torch.Tensor.dtype.__get__)
def get_dtype(t: CategoricalTensor):
    return t.tensor.dtype